/*
# Copyright (C) 1999 Scott Dattalo
# 
# This file is part of gnupic project.
# 
# vc is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
# 
# vc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with vc; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.  */

/*
      .COD files

      Copyright (c) 1989, 1994, 1996                                                        
                                                                                
      Walter Banks
      Byte Craft Limited
      421 King Street North
      Waterloo, Ontario
      CANADA N2J 4E4
      (519) 888 6911
      EMAIL: walter@bytecraft.com

      Please note that Byte Craft did not write this code and consequently
      is not responible in any way for what it may do. Byte Craft designed
      the structure of .COD files and so deserve some recognition here...
*/

/*
  This file is basically the c version of Byte Craft's pascal version.
  If you have access to that file you'll notice that many of the variable
  names and structures are identical. In fact, I even ran p2c to get some
  of the information. Having said that though, you'll also see a number
  of differences. These differences are primarily to make the functionality
  of this program easily integratable (sp?) into gpsim and gpasm; the 
  gnupic simulator and assembler.

  The information about the detailed structure of .cod files is freely
  available from Walter Banks at Byte Craft. The information about .cod
  files that you find in here is provided just to give you an idea of
  what the program is attempting to accomplish. If you have questions
  about the details of .cod files I probably don't know the answers.
  Ask Walter.

*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "cod.h"

#define BUFFER_LENGTH 256
#define BLOCK_SIZE    512

FILE *codefile;  
char filename[BUFFER_LENGTH];
char temp[BLOCK_SIZE];
char *source_file_names[20];
FILE *source_files[20];

int number_of_source_files = 0;
#define VIEW_CODE_VERSION_STRING "vc 0.0.1"

char directory_block_data[BLOCK_SIZE];
static char * SymbolType4[154] = {
  "a_reg          ", "x_reg          ", "c_short        ", "c_long         ",
  "c_ushort       ", "c_ulong        ", "c_pointer      ", "c_upointer     ",
  "table          ", "m_byte         ", "m_boolean      ", "m_index        ",
  "byte_array     ", "u_byte_array   ", "word_array     ", "u_word_array   ",
  "func_void_none ", "func_void_byte ", "funcVoidTwobyte", "func_void_long ",
  "func_void_undef", "func_int_none  ", "func_int_byte  ", "func_int_twobyt",
  "func_int_long  ", "func_int_undef ", "func_long_none ", "func_long_byte ",
  "funcLongTwobyte", "func_long_long ", "func_long_undef", "pfun_void_none ",
  "pfun_void_byte ", "pfunVoidTwobyte", "pfun_void_long ", "pfun_void_undef",
  "pfun_int_none  ", "pfun_int_byte  ", "pfunIntTwobyte ", "pfun_int_long  ",
  "pfun_int_undef ", "pfun_long_none ", "pfun_long_byte ", "pfun_long_twoby",
  "pfun_long_long ", "pfun_long_undef", "address        ", "constant       ",
  "bad_und        ", "br_und         ", "upper_und      ", "byte_und       ",
  "add_und        ", "m_keyword      ", "add_mi_und     ", "vector         ",
  "port_w         ", "port_r         ", "port_rw        ", "port_rmw       ",
  "endif          ", "exundef        ", "macro_t        ", "macro_s        ",
  "macro_a        ", "macro_c        ", "c_keyword      ", "void           ",
  "c_enum         ", "c_typedef1     ", "c_utypedef1    ", "c_typedef2     ",
  "c_utypedef2    ", "cp_typedef1    ", "cp_utypedef1   ", "cp_typedef2    ",
  "cp_utypedef2   ", "c_struct       ", "i_struct       ", "l_const        ",
  "s_short        ", "s_ushort       ", "s_long         ", "s_ulong        ",
  "sa_short       ", "sa_ushort      ", "sa_long        ", "sa_ulong       ",
  "sp_short       ", "sp_ushort      ", "sp_long        ", "sp_ulong       ",
  "FixedPointer   ", "PointerFunction", "cc_reg         ", "PtrF_void_none ",
  "PtrF_void_byte ", "PtrF_void_twobyt", "PtrF_void_long ", "PtrF_void_undef",
  "PtrF_int_none  ", "PtrF_int_byte  ", "PtrF_int_twobyte", "PtrF_int_long  ",
  "PtrF_int_undef ", "PtrF_long_none ", "PtrF_long_byte ",
  "PtrF_long_twobyte", "PtrF_long_long ", "PtrF_long_undef",
  "PfAR_void_none ", "PfAR_void_byte ", "PfAR_void_twobyte",
  "PfAR_void_long ", "PfAR_void_undef", "PfAR_int_none  ", "PfAR_int_byte  ",
  "PfAR_int_twobyte", "PfAR_int_long  ", "PfAR_int_undef ", "PfAR_long_none ",
  "PfAR_long_byte ", "PfAR_long_twobyte", "PfAR_long_long ",
  "PfAR_long_undef", "FWDlit         ", "Pfunc          ", "mgoto          ",
  "mcgoto         ", "mcgoto2        ", "mcgoto3        ", "mcgoto4        ",
  "mcgoto74       ", "mcgoto17       ", "mccall17       ", "mcall          ",
  "mc_call        ", "res_word       ", "local          ", "unknown        ",
  "varlabel       ", "external       ", "global         ", "segment        ",
  "Bankaddr       ", "bit_0          ", "bit_1          ", "bit_2          ",
  "bit_3          ", "bit_4          ", "bit_5          ", "bit_6          ",
  "bit_7          ", "debug          "
};


void dump_hex(char *chunk, int length)
{

  int i,j,all_zero_line;

  printf("\n");

  i = 0;
  do {

    for(j=0, all_zero_line=1; j<16; j++)
      if(chunk[i+j]) all_zero_line = 0;

    if(all_zero_line)
      i+=16;
    else {
      printf("\n%04x:  ",i);

      for(j=0; j<16; j++)
	printf("%02x ",(unsigned char)chunk[i+j]);

      printf("   ");

      for(j=0; j<16; j++)
	if(isprint(chunk[i+j]))
	  putchar(chunk[i+j]);
	else
	  putchar('.');

      i+=16;
    }

  }while (i<length);

  printf("\n");

}



char * substr(char *a, char *b, int n)
{
  *a = 0;
  return strncat(a,b,n);

}

// a little utility routine that's not actually used in the
// program, but was useful in computing offsets into structures.

int ptr_offset(char *a, char *b)
{

  return(  ( (int)a) - ((int)b) );

}

/*
 * get_short_int, put_short_int, and put_int
 * utility routines for accessing the cod blocks (char *)
 * using the proper endianess. (Using a union would be
 * cleaner, but less portable.
 */

unsigned short get_short_int( char * buff)
{
  return ( (unsigned char)buff[0] + ((unsigned char)buff[1] << 8));
}

void put_short_int(char * buff, int value)
{

  buff[0] = value & 0xff;
  buff[1] = (value >> 8) & 0xff;
}

void put_int(char * buff, int value)
{

  buff[0] = value & 0xff;
  buff[1] = (value >> 8) & 0xff;
  buff[2] = (value >> 16) & 0xff;
  buff[3] = (value >> 24) & 0xff;
}

// reverse the endian type (big to little or little to big)
// note: there's probably a wonderful macro nestled deep in
// the bowels of the os that nicely handles this conversion...

unsigned int endian_swap(unsigned int a)
{

  return
    ( (a>>24) & 0x000000ff) | 
    ( (a>>8)  & 0x0000ff00) |
    ( (a<<8)  & 0x00ff0000) |
    ( (a<<24) & 0xff000000);

}


void directory_block(char *block)
{
  char temp_buf[256];

  Directory *dir;

  dir = (Directory *)block;

  printf("directory block \n");

  printf("COD file version  %d\n",
	 get_short_int(&block[COD_DIR_CODTYPE]));
  printf("Source file       %s\n",
	 &block[COD_DIR_SOURCE]);
  printf("Date              %s\n",
	 substr(temp_buf,&block[COD_DIR_DATE],7));
  printf("Time              %2d:%2d\n",
	 get_short_int(&block[COD_DIR_TIME]) / 100, 
	 get_short_int(&block[COD_DIR_DATE]) % 100);
  printf("Compiler version  %s\n",
	 substr(temp_buf,&block[COD_DIR_VERSION],19));
  printf("Compiler          %s\n",
	 substr(temp_buf,&block[COD_DIR_COMPILER],12));
  printf("Notice            %s\n",
	 substr(temp_buf,&block[COD_DIR_NOTICE],64));
  printf("Processor         %s\n",
	 substr(temp_buf,&block[COD_DIR_PROCESSOR],8));

  printf("Bytes for address: %d\n",
	 get_short_int(&block[COD_DIR_ADDRSIZE]+2));
  printf("High word of 64k address %04x\n",dir->HighAddr);

  printf("Short symbol table start block:  0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_SYMTAB]),
	 get_short_int(&block[COD_DIR_SYMTAB+2]));
  printf("Long symbol table start block:   0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_LSYMTAB]),
	 get_short_int(&block[COD_DIR_LSYMTAB+2]));
  printf("File name table start block:     0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_NAMTAB]),
	 get_short_int(&block[COD_DIR_NAMTAB+2]));
  printf("Source info table start block:   0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_LSTTAB]),
	 get_short_int(&block[COD_DIR_LSTTAB+2]));
  printf("Rom table start block:           0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_MEMMAP]),
	 get_short_int(&block[COD_DIR_MEMMAP+2]));
  printf("Local scope table start block:   0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_LOCALVAR]),
	 get_short_int(&block[COD_DIR_LOCALVAR+2]));
  printf("Debug messages start block:      0x%04x  end block:  0x%04x\n",
	 get_short_int(&block[COD_DIR_MESSTAB]),
	 get_short_int(&block[COD_DIR_MESSTAB+2]));

  printf("\nNext directory block");
  if(get_short_int(&block[COD_DIR_NEXTDIR]))
    printf(":  %d\n",get_short_int(&block[COD_DIR_NEXTDIR]));
  else
    printf(" is empty\n");

  /*
    Uncomment this section to get the address offsets within the directory structure
    NOTE: the offsets are compiler dependent. For gcc, use the -fpack-struct
    option to pack the structures with no gaps.

  printf("notice_strlen address %d\n",ptr_offset((char *)&dir->notice_strlen,block) );
  printf("notice address %d\n",ptr_offset((char *)&dir->notice,block) );
  printf("symtab address %d\n",ptr_offset((char *)&dir->symtab,block) );
  printf("processor address %d\n",ptr_offset((char *)&dir->processor,block) );
  printf("AddrSize address %d\n",ptr_offset((char *)&dir->AddrSize,block) );
  printf("HighAddr address %d\n",ptr_offset((char *)&dir->HighAddr,block) );
  printf("NextDir address %d\n",ptr_offset((char *)&dir->NextDir,block) );
  printf("memmap address %d\n",ptr_offset((char *)&dir->MemMapOFS,block) );
  printf("lsymtab address %d\n",ptr_offset((char *)&dir->Lsymtab,block) );
  */


}

void read_block(char * block, int block_number)
{

  fseek(codefile, block_number * BLOCK_SIZE, SEEK_SET);
  fread(block, BLOCK_SIZE, 1, codefile);
}

/*---------------------------------------------*/
/* 
 * ROM usage information
 */

void dump_memmap( void)
{
  typedef struct _REC_maptab {
    unsigned short start, last;
  } _REC_maptab;

  _REC_maptab *mt;
  int i,j,start_block,end_block;

  start_block = get_short_int(&directory_block_data[COD_DIR_MEMMAP]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_MEMMAP+2]);

    printf("\n\nROM Usage\n");
    printf("--------------------------------\n");

    for(j=start_block; j<=end_block; j++) {

      read_block(temp, j);

      mt = (_REC_maptab *)temp;

      for(i=0; i< 128; i++)
	if( !((mt[i].start == 0) && (mt[i].last == 0) ))
	  printf("using ROM 0x%04x to 0x%04x\n",mt[i].start/2,mt[i].last/2);
    }
  } else
    printf("    No ROM usage information available.\n");

}

/*---------------------------------------------*/
/*
 * Dump all of the machine code in the .cod file 
 */

void dump_code(void)
{

  int i,j,k,all_zero_line,index;

  dump_memmap();

  printf("\n\nFormatted Code Dump\n");
  printf("-------------------\n");

  for (k = 0; k <= 127; k++) {
    index = get_short_int(&directory_block_data[2*(COD_DIR_CODE + k)]);
    if (index != 0) {
      read_block(temp, index);

      printf("\n");

      i = 0;
      do {

	for(j=0, all_zero_line=1; j<8; j++)
	  if( get_short_int(&temp[(i+j)*2]) ) all_zero_line = 0;

	if(all_zero_line)
	  i+=8;
	else {
	  printf("\n%06x:  ",i+k*256);

	  for(j=0; j<8; j++)
	    printf("%04x ",get_short_int(&temp[2*i++]));
	}

      }while (i<BLOCK_SIZE/2);

      printf("\n");
    }
  }
}

/*---------------------------------------------*/
/*
 * Dump all of the (short) Symbol Table stuff in the .cod file 
 */


void dump_symbols( void )
{
#define SYMBOL_SIZE  16
#define SYMBOLS_PER_BLOCK BLOCK_SIZE/SYMBOL_SIZE

#define SR_LEN       0
#define SR_NAME      1
#define SR_TYPE      13
#define SR_VALUE     14

  /* The symbol record structure
typedef struct symbol_record {
  char name_strlen;
  char name[12];
  uchar stype;
  unsigned short svalue;
} symbol_record;
  */

  int i,j,start_block,end_block;
  char b[16];

  start_block = get_short_int(&directory_block_data[COD_DIR_SYMTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_SYMTAB+2]);

    printf("\nSymbol Table Information\n");
    printf("------------------------\n\n");

    for(j=start_block; j<=end_block; j++) {

      read_block(temp, j);

      for(i=0; i<SYMBOLS_PER_BLOCK; i++) {

	if(temp[i*SYMBOL_SIZE + SR_NAME])
	  printf("%s = %d, type = %s\n",
		 substr(b,&temp[i*SYMBOL_SIZE + SR_NAME],12), 
		 get_short_int(&temp[i*SYMBOL_SIZE + SR_VALUE]), 
		 SymbolType4[(unsigned char)temp[i*SYMBOL_SIZE + SR_TYPE]]
		 );
      }
    }
  }
  else
    printf("No symbol table info\n");

}

/*---------------------------------------------*/

/*
 * Dump all of the Long Symbol Table stuff in the .cod file 
 */

void dump_lsymbols( void )
{
  char *s,length;
  short type;
  int i,j,start_block,end_block, value;
  char b[256];

  start_block = get_short_int(&directory_block_data[COD_DIR_LSYMTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_LSYMTAB+2]);

    printf("\nLong Symbol Table Information\n");
    printf("------------------------\n\n");

    for(j=start_block; j<=end_block; j++) {

      read_block(temp, j);

      for(i=0; i<BLOCK_SIZE;) {
	s =  &temp[i];

	if(*s==0)
	  break;

	length = *s;
	type  = *(short *)&s[length+1];
	if(type>128)
	  type = 0;
	value = *(int *)&s[length+3];
	value = endian_swap(value);

	printf("%s = %x, type = %s\n",
	       substr(b,&s[1],length),
	       value,
	       SymbolType4[type]);
	i += (length + 7);
      }
    }
  }else
      printf("No long symbol table info\n");

}

/*---------------------------------------------*/
/*
 * Source files
 */

void dump_source_files( void )
{
#define FILE_SIZE  64
#define FILES_PER_BLOCK BLOCK_SIZE/FILE_SIZE

  int i,j,start_block,end_block,offset;
  char b[FILE_SIZE];

  start_block = get_short_int(&directory_block_data[COD_DIR_NAMTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_NAMTAB+2]);

    printf("\nSource File Information\n");
    printf("------------------------\n\n");

    for(j=start_block; j<=end_block; j++) {

      read_block(temp, j);

      for(i=0; i<FILES_PER_BLOCK; i++) {

	offset = i*FILE_SIZE;
	substr(b,&temp[offset+1],FILE_SIZE);

	if(temp[offset]) {
	  source_file_names[number_of_source_files] = (char *)malloc(strlen(b));
	  strcpy(source_file_names[number_of_source_files],b);
	  printf("%s\n",source_file_names[number_of_source_files]);
	  source_files[number_of_source_files] = 
	    fopen(source_file_names[number_of_source_files],"r");
	  number_of_source_files++;
	}
      }
    }
  }else
    printf("No source file info\n");
}

char * smod_flags(int smod)
{
  static char f[9];

  f[0] = (smod & 0x80) ? 'C':'.';
  f[1] = (smod & 0x40) ? 'F':'.';
  f[2] = (smod & 0x20) ? 'I':'.';
  f[3] = (smod & 0x10) ? 'D':'.';
  f[4] = (smod & 0x08) ? 'C':'.';
  f[5] = (smod & 0x04) ? 'L':'.';
  f[6] = (smod & 0x02) ? 'N':'.';
  f[7] = (smod & 0x01) ? 'A':'.';
  f[8] = 0;

  return(f);
}

/*---------------------------------------------*/
/*
 * Line number info from the source files
 */

void dump_line_symbols(void)
{
  static int lst_line_number = 1;
  char buf[256];
  int i,j,start_block,end_block;

  LineSymbol *ls;


  start_block = get_short_int(&directory_block_data[COD_DIR_LSTTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_LSTTAB+2]);

    printf("\n\nLine Number Information\n");
    printf(" LstLn  SrcLn  Addr   Flags         FileName\n");
    printf(" -----  -----  ----   -----------   ---------------------------------------\n");

    for(j=start_block; j<=end_block; j++) {

      read_block(temp, j);

      ls = (LineSymbol *) temp;

      for(i=0; i<84; i++) {

	if((ls[i].smod & 4) == 0)
	  printf(" %5d  %5d %5d   %2x %s   %-50s\n",
		 lst_line_number++,
		 ls[i].sline,
		 ls[i].sloc,
		 ls[i].smod,
		 smod_flags(ls[i].smod),
		 source_file_names[ls[i].sfile]);
	if(source_files[ls[i].sfile]) {
	  fgets(buf, 256, source_files[ls[i].sfile]);
	  printf("%s",buf);
	}
      }
    }
  } else
    printf("No line number info\n");
}
/*---------------------------------------------*/
/*
 * Debug Message area
 */

void dump_message_area(void)
{
  char DebugType,DebugMessage[64];

  int i,j,start_block,end_block;
  int laddress;

  j=0;

  start_block = get_short_int(&directory_block_data[COD_DIR_MESSTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_MESSTAB+2]);

    printf("\n\nDebug Message area\n");
    printf("------------------\n\n");

    for(i=start_block; i<=end_block; i++) {
      read_block(temp, i);

      while (j < 504) {
	if (temp[j + 6] == 0) {
	  j = 512;   /* done */
	  break;
	}

	laddress = endian_swap(*(int *)&temp[j]);

	DebugType = temp[j++];

	substr(DebugMessage,&temp[j],64);
	j += strlen(DebugMessage);

	printf("%8x %2d %s\n",laddress, DebugType, DebugMessage);
      }
    }
  } else
    printf("    No Debug Message information available.\n");
    
}
/*---------------------------------------------*/
/* 
 * Display the local symbol table information
 */

void dump_local_vars(void)
{
#define SYMBOL_SIZE  16
#define SYMBOLS_PER_BLOCK BLOCK_SIZE/SYMBOL_SIZE

  typedef struct symbol_record {
    char name_strlen;
    char name[12];         /* Not Symbol Length because */
    unsigned char stype;   /* this record is 16 bytes   */
    unsigned short svalue;
  } symbol_record;

  typedef struct scope_head {
    char name_strlen;
    char name[7];
    unsigned int start;
    unsigned int stop;
  } scope_head;

  symbol_record *sr;
  scope_head  *sh;

  int i,j,start_block,end_block;

  j=0;

  start_block = get_short_int(&directory_block_data[COD_DIR_LOCALVAR]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_LOCALVAR+2]);

    printf("\n\nLocal Symbol Scoping Information\n");
    printf("--------------------------------\n");

    for(i=start_block; i<=end_block; i++) {
      read_block(temp, i);

      for(i=0; i<SYMBOLS_PER_BLOCK; i++) {
	sh = (scope_head *)&temp[i*SYMBOL_SIZE];

	if(sh->name[0]) {
	  if( 0 == strncmp(sh->name,"__LOCAL",7)) {
	    printf("Local symbols between %06x and %06x:  ",sh->start,sh->stop);
	  } else {
	    sr = (symbol_record *)sh;
	    printf("%s = 0x%x, type = %s\n",sr->name, sr->svalue, SymbolType4[sr->stype]);
	  }
	}
      }
    }
  }else
    printf("    No local variable scoping info available.\n");

}



/*---------------------------------------------*/

int main(int argc, char *argv[])
{

  unsigned int buffer_size;
  int c,usage=0;
  int display_flags;
  Directory *dir;

#define DISPLAY_NOTHING 0
#define DISPLAY_DIR     1
#define DISPLAY_SYM     2
#define DISPLAY_ROM     4
#define DISPLAY_SRC     8
#define DISPLAY_MESS    16
#define DISPLAY_ALL     0xff


  printf("Copyright (c) 1989, 1994 Byte Craft Limited Waterloo, \n");
  printf("Ontario CANADA. All rights reserved.\n");

  display_flags = DISPLAY_NOTHING;

  while ((c = getopt(argc, argv, "h?adsrlmv")) != EOF) {
    switch (c) {
    case '?':
    case 'h':
      usage = 1;
      break;
    case 'a':
      display_flags = DISPLAY_ALL;
      break;
    case 'd':
      display_flags |= DISPLAY_DIR;
      break;
    case 's':
      display_flags |= DISPLAY_SYM;
      break;
    case 'r':
      display_flags |= DISPLAY_ROM;
      break;
    case 'l':
      display_flags |= DISPLAY_SRC;
      break;
    case 'm':
      display_flags |= DISPLAY_MESS;
      break;
    case 'v':
      fprintf(stderr, "%s\n", VIEW_CODE_VERSION_STRING);
      exit(0);
    }
    if (usage)
      break;
  }
  
  if (optind < argc)
    strcpy(filename, argv[optind]);
  else
    usage = 1;

  if(display_flags == DISPLAY_NOTHING)
    display_flags = DISPLAY_ALL;


  if (usage) {
      printf("\n.cod file decoder\n");
      printf("usage: vc <options> <filename>\n");
      printf("Where <options> are:\n");
      printf(" -a        Display all information in .cod file\n");
      printf(" -d        Display directory header\n");
      printf(" -s        Display symbols\n");
      printf(" -h        Show this usage message \n");
      printf(" -r        Display rom\n");
      printf(" -l        Display source listing cross\n");
      printf(" -m        Display debug message area\n");
      printf(" -v        Show version\n");
      printf("\n");
      exit(0);
  }

  codefile = fopen(filename,"r");
  if(codefile == NULL) {
    printf("Unable to open %s\n",filename);
    exit(1);
  }


  /* Start off by reading the directory block */

  buffer_size = fread(directory_block_data, 1, BLOCK_SIZE, codefile);

  if(display_flags & DISPLAY_DIR)
    directory_block(directory_block_data);

  dir = (Directory *)directory_block_data;

  if(display_flags & DISPLAY_ROM)
    dump_code();

  if(display_flags & DISPLAY_SYM) {
    dump_symbols();
    dump_lsymbols();
    dump_local_vars();
  }


  dump_source_files();


  if(display_flags & DISPLAY_SRC)
    dump_line_symbols();

  if(display_flags & DISPLAY_MESS)
    dump_message_area();

  return(0);
}
